@@ -2,7 +2,7 @@
 #'
 #' Read orderings from `.soc`, `.soi`, `.toc` or `.toi` files storing
 #' ordinal preference data format as defined by
-#' \href{https://www.preflib.org/}{\{PrefLib\}: A Library for Preferences}
+#' \href{https://preflib.github.io/PrefLib-Jekyll/}{\{PrefLib\}: A Library for Preferences}
 #' into a `preferences` object.
 #'
 #' Note that PrefLib refers to the items being ordered by "alternatives".
@@ -16,17 +16,15 @@
 #' }
 #'
 #' The numerically coded orderings and their frequencies are read into a
-#' data frame, storing the item names as an attribute. The
-#' `as.aggregated_preferences` method converts these to an
-#' [`aggregated_preferences`][aggregate.preferences] object with the
-#' items labelled by name.
+#' tibble, storing all original metadata in a "preflib" attribute.
 #'
 #' A PrefLib file may be corrupt, in the sense that the ordered alternatives do
-#' not match their names. In this case, the file can be read in as a data
-#' frame (with a warning), but `as.aggregated_preferences` will throw an error.
+#' not match their names. In this case, the file will still be read, but with a
+#' warning.
 #'
-#' @return An [`aggregated_preferences`][aggregate.preferences] object
-#' containing the PrefLib data.
+#' @return A tibble with two columns: `preferences` and `frequency`. The
+#' `preferences` column contains all the preferential orderings in the file, and
+#' the `frequency` column the relative frequency of this selection.
 #'
 #' @param file A preferential data file, conventionally with extension `.soc`,
 #' `.soi`, `.toc` or `.toi` according to data type.
@@ -51,23 +49,42 @@
 #'
 #' \donttest{
 #' # strict complete orderings of four films on Netflix
-#' netflix <- read_preflib("netflix/00004-00000138.soc", from_preflib = TRUE)
+#' netflix <- read_preflib("00004 - netflix/00004-00000138.soc", from_preflib = TRUE)
 #' head(netflix)
-#' names(netflix$preferences)
+#' levels(netflix$preferences)
 #'
 #' # strict incomplete orderings of 6 random cities from 36 in total
-#' cities <- read_preflib("cities/00034-00000001.soi", from_preflib = TRUE)
+#' cities <- read_preflib("00034 - cities/00034-00000001.soi", from_preflib = TRUE)
 #' }
-#' @importFrom utils read.csv
+#' @importFrom utils read.csv URLencode
+#' @importFrom tibble tibble
 #' @export
 read_preflib <- function(file,
                          from_preflib = FALSE,
-                         preflib_url = "https://www.preflib.org/static/data") {
+                         preflib_url = "https://raw.githubusercontent.com/PrefLib/PrefLib-Data/main/datasets/") {
   if (from_preflib) {
-    file <- file.path(preflib_url, file)
+    tryCatch(
+      {
+        # Check if the URL is accessible before proceeding
+        lines <- readLines(paste0(preflib_url, URLencode(file)))
+      },
+      warning = function(w) {
+        stop(
+          "Could not access PrefLib resource.\n",
+          conditionMessage(w)
+        )
+      },
+      error = function(e) {
+        stop(
+          "Could not access PrefLib resource.\n",
+          conditionMessage(e)
+        )
+      }
+    )
+  } else {
+    # Read local file into memory
+    lines <- readLines(file)
   }
-  # Read the file into memory
-  lines <- readLines(file)
 
   # Filter the header lines and convert them to key:value attributes
   header_lines <- sub("# ", "", grep("^#", lines, value = TRUE))
@@ -172,52 +189,57 @@ read_preflib <- function(file,
     as.integer,
     how = "replace"
   )
-  names(preferences) <- c(
-    "Frequency",
-    paste("Rank", seq_len(ncol(preferences) - 1L))
-  )
-  frequencies <- preferences[, 1L]
-  preferences <- preferences[, -1L]
-  preferences <- as.preferences(
-    preferences,
-    "ordering",
-    item_names = alternative_names
-  )
-  aggregated_prefs <- aggregate(preferences, frequencies = frequencies)
-  attr(aggregated_prefs, "preflib") <- preflib_attributes
+  frequency <- preferences[, 1L]
+  # Convert orderings to vctrs format
+  preferences <- preferences[, -1L] |>
+    apply(
+      1L,
+      \(ord) {
+        lapply(
+          seq_along(ord),
+          \(r) cbind(ord[[r]], rep(r, length(ord[[r]])))
+        ) |>
+          do.call(what = rbind)
+      },
+      simplify = FALSE
+    ) |>
+    .vctr_preferences(item_names = alternative_names)
+
+  output_tibble <- tibble(preferences = preferences, frequency = frequency)
+  attr(output_tibble, "preflib") <- preflib_attributes
 
   # Ensure we have the expected number of unique and total orderings.
   n_unique_orders <- as.integer(preflib_attributes[["NUMBER UNIQUE ORDERS"]])
   n_voters <- as.integer(preflib_attributes[["NUMBER VOTERS"]])
-  if (length(aggregated_prefs$preferences) != n_unique_orders) {
+  if (length(preferences) != n_unique_orders) {
     warning(
       "Expected ",
       n_unique_orders,
       " unique orderings but only ",
-      length(aggregated_prefs$preferences),
+      length(preferences),
       " were recovered when reading the PrefLib datafile.",
       " The file may be corrupt."
     )
   }
-  if (sum(aggregated_prefs$frequencies) != n_voters) {
+  if (sum(frequency) != n_voters) {
     warning(
       "Expected ",
       n_voters,
       " total orderings but only ",
-      sum(aggregated_prefs$frequencies),
+      sum(frequency),
       " were recovered when reading the PrefLib datafile.",
       " The file may be corrupt."
     )
   }
 
-  return(aggregated_prefs)
+  return(output_tibble)
 }
 
 #' Write Ordinal Preference Data to PrefLib Formats
 #'
 #' Write `preferences` to `.soc`, `.soi`, `.toc` or `.toi` file types, as
 #' defined by the PrefLib specification:
-#' \href{https://www.preflib.org/}{\{PrefLib\}: A Library for Preferences}.
+#' \href{https://preflib.github.io/PrefLib-Jekyll/}{\{PrefLib\}: A Library for Preferences}.
 #'
 #' @details The file types supported are
 #' \describe{
@@ -285,11 +307,16 @@ read_preflib <- function(file,
 #' The "alternatives" in the output file will be the same as the "items" in the
 #' `aggregated_preferences` object.
 #'
-#' @param x An `aggregated_preferences` object to write to file. If `x` is of a
-#' different class, it attempts to coerce `x` into an `aggregated_preferences`
-#' object via `as.aggregated_preferences()`.
+#' @param x A `preferences` object or a `tibble` with a `preferences`-typed
+#' column to write to file.
 #' @param file Either a character string naming the a file or a writeable,
 #' open connection. The empty string `""` will write to stdout.
+#' @param preferences_col <[`tidy-select`][dplyr::dplyr_tidy_select]> When `x` is a
+#' `tibble`, the column containing the preferences to be written to file.
+#' If not provided and `x` is a `tibble`, then
+#' @param frequency_col <[`tidy-select`][dplyr::dplyr_tidy_select]> When `x` is a
+#' `tibble`, the column containing the frequency of the preferences. If not
+#' provided, each row is considered to be observed a single time.
 #' @param title The title of the data file, for instance the name of the
 #' election represented in the data file. If not provided, we check for the
 #' presence of `attr(x, "preflib")`, and if it exists we check for `TITLE`.
@@ -318,6 +345,8 @@ read_preflib <- function(file,
 #' @export
 write_preflib <- function(x, # nolint: cyclocomp_linter
                           file = "",
+                          preferences_col = NULL,
+                          frequency_col = NULL,
                           title = NULL,
                           publication_date = NULL,
                           modification_type = NULL,
@@ -430,8 +459,11 @@ write_preflib <- function(x, # nolint: cyclocomp_linter
     stop("'file' must be a character string.")
   }
 
-  # Coerce into aggregated_preferences
-  x <- as.aggregated_preferences(x)
+  x <- .validate_preferences_frequencies(
+    x,
+    {{ preferences_col }},
+    {{ frequency_col }}
+  )
 
   # Prepare lines for file
   lines <- c(
@@ -439,15 +471,15 @@ write_preflib <- function(x, # nolint: cyclocomp_linter
     paste("#", "FILE NAME:", file_name),
     paste("#", "TITLE:", title),
     paste("#", "DESCRIPTION:", description),
-    paste("#", "DATA TYPE:", preftype(x$preferences)),
+    paste("#", "DATA TYPE:", pref_type(x$preferences)),
     paste("#", "MODIFICATION TYPE:", modification_type),
     paste("#", "RELATES TO:", paste(relates_to, collapse = ",")),
     paste("#", "RELATED FILES:", paste(related_files, collapse = ",")),
     paste("#", "PUBLICATION DATE:", publication_date),
     paste("#", "MODIFICATION DATE:", modification_date),
     # Derived metadata
-    paste("#", "NUMBER ALTERNATIVES:", length(names(x$preferences))),
-    paste("#", "NUMBER VOTERS:", sum(x$frequencies)),
+    paste("#", "NUMBER ALTERNATIVES:", nlevels(x$preferences)),
+    paste("#", "NUMBER VOTERS:", sum(x$frequency)),
     paste("#", "NUMBER UNIQUE ORDERS:", length(x$preferences))
   )
 
@@ -461,20 +493,20 @@ write_preflib <- function(x, # nolint: cyclocomp_linter
   )
 
   # Format orders as strings
-  orderings <- apply(
-    as.matrix(x$preferences),
-    1L,
-    function(item_ranks) {
-      paste0(
-        sapply(
-          seq_len(max(na.omit(item_ranks))),
-          function(ri) fmt_eql_items(which(ri == item_ranks))
-        ),
-        collapse = ","
-      )
-    }
-  )
-  lines <- c(lines, paste0(x$frequencies, ": ", orderings))
+
+  ordering_str <- vctrs::vec_data(x$preferences) |>
+    sapply(
+      \(o) {
+        paste0(
+          sapply(
+            split(o[, 1L], o[, 2L]),
+            fmt_eql_items # Formats ties ~ {1,2,3,...}
+          ),
+          collapse = ","
+        )
+      }
+    )
+  lines <- c(lines, paste0(x$frequency, ": ", ordering_str))
 
   writeLines(lines, file, sep = "\n")
   if (file_name != "NA") {
@@ -482,11 +514,11 @@ write_preflib <- function(x, # nolint: cyclocomp_linter
   }
 }
 
-# Helper function for formatting list of items at equal rank
+# Helper function for formatting list of items at equal rank ~ {1,2,3,...}
 fmt_eql_items <- function(items) {
   if (length(items) > 1L) {
     paste0("{", paste0(items, collapse = ","), "}")
   } else {
-    as.character(items)
+    items
   }
 }
